@@ -1,6 +1,6 @@
 import re
 from enum import Enum
-from typing import Generic, List
+from typing import Generic, List, Set
 
 from jsonschema import Draft7Validator, SchemaError
 
@@ -59,7 +59,7 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
         >>> from mistral_common.protocol.instruct.messages import UserMessage, AssistantMessage
         >>> validator = MistralRequestValidator()
         >>> messages = [UserMessage(content="Hello how are you ?")]
-        >>> validator.validate_messages(messages)
+        >>> validator.validate_messages(messages, False)
     """
 
     _allow_tool_call_and_content: bool = False
@@ -72,19 +72,20 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
         """
         self._mode = mode
 
-    def validate_messages(self, messages: List[UATS]) -> None:
+    def validate_messages(self, messages: List[UATS], continue_final_message: bool) -> None:
         r"""Validates the list of messages.
 
         Args:
             messages: The list of messages to validate.
+            continue_final_message: Whether to continue the final message.
 
         Examples:
             >>> from mistral_common.protocol.instruct.messages import UserMessage, AssistantMessage
             >>> validator = MistralRequestValidator()
             >>> messages = [AssistantMessage(content="Hi"), UserMessage(content="Hello")]
-            >>> validator.validate_messages(messages)
+            >>> validator.validate_messages(messages, False)
         """
-        self._validate_message_list_structure(messages)
+        self._validate_message_list_structure(messages, continue_final_message=continue_final_message)
         self._validate_message_list_content(messages)
 
     def validate_request(self, request: ChatCompletionRequest) -> ChatCompletionRequest[UATS]:
@@ -108,7 +109,7 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
                 raise InvalidRequestException("Model name parameter is required for serving mode")
 
         # Validate the messages
-        self.validate_messages(request.messages)
+        self.validate_messages(request.messages, continue_final_message=request.continue_final_message)
 
         # Validate the tools
         self._validate_tools(request.tools or [])
@@ -209,7 +210,7 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
         if message.prefix:
             if not is_last_message:
                 raise InvalidAssistantMessageException("Assistant message with prefix True must be last message")
-            # note : we already validate that assistant messsage has content 3 lines up.
+            # note : we already validate that assistant message has content 3 lines up.
 
     def _validate_tool_calls_followed_by_tool_messages(self, messages: List[UATS]) -> None:
         """
@@ -268,7 +269,7 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
 
             previous_role = current_role
 
-    def _validate_last_message(self, message: UATS) -> None:
+    def _validate_last_message(self, message: UATS, continue_final_message: bool) -> None:
         # The last message must be a user or tool message in serving mode or an assistant message in finetuning mode
         last_message_role = message.role
         if self._mode == ValidationMode.finetuning:
@@ -276,16 +277,23 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
                 raise InvalidMessageStructureException(
                     f"Expected last role Assistant for finetuning but got {last_message_role}"
                 )
+            if continue_final_message:
+                raise InvalidMessageStructureException("Cannot continue final message in finetuning mode")
         else:
-            bad_assistant = isinstance(message, AssistantMessage) and not message.prefix
+            bad_assistant = isinstance(message, AssistantMessage) and not message.prefix and not continue_final_message
             bad_role = message.role not in {Roles.user, Roles.tool}
             if bad_assistant and bad_role:
                 raise InvalidMessageStructureException(
-                    f"Expected last role User or Tool (or Assistant with prefix True) for serving"
-                    f" but got {last_message_role}"
+                    f"Expected last role User or Tool (or Assistant with prefix or continue_final_message set to True) "
+                    f"for serving but got {last_message_role}"
+                )
+            elif continue_final_message and (last_message_role != Roles.assistant or message.prefix):
+                raise InvalidMessageStructureException(
+                    f"Expected last role Assistant with prefix False for serving with continue_final_message set to "
+                    f"True but got {last_message_role}"
                 )
 
-    def _validate_message_list_structure(self, messages: List[UATS]) -> None:
+    def _validate_message_list_structure(self, messages: List[UATS], continue_final_message: bool) -> None:
         """
         Validates the structure of the list of messages
 
@@ -302,7 +310,7 @@ class MistralRequestValidator(Generic[UserMessageType, AssistantMessageType, Too
 
         # Always check the last message if in fine-tuning mode
         if self._mode == ValidationMode.finetuning or len(messages) > 1:
-            self._validate_last_message(messages[-1])
+            self._validate_last_message(messages[-1], continue_final_message=continue_final_message)
 
         self._validate_message_order(messages)
         self._validate_tool_calls_followed_by_tool_messages(messages)
@@ -373,8 +381,8 @@ class MistralRequestValidatorV3(MistralRequestValidator):
 
         self._validate_function_call(tool_call.function)
 
-    def _validate_last_message(self, message: UATS) -> None:
-        super()._validate_last_message(message)
+    def _validate_last_message(self, message: UATS, continue_final_message: bool) -> None:
+        super()._validate_last_message(message, continue_final_message)
 
         if self._mode == ValidationMode.finetuning:
             # in finetuning mode it has to be an assistant message
@@ -394,3 +402,52 @@ class MistralRequestValidatorV5(MistralRequestValidatorV3):
     """
 
     _allow_tool_call_and_content: bool = True
+
+
+class MistralRequestValidatorV13(MistralRequestValidatorV5):
+    def _validate_tool_calls_followed_by_tool_messages(self, messages: List[UATS]) -> None:
+        """
+        Checks:
+        - That the number and ids of tool calls and tool messages are the same
+        - That the tool calls are followed by tool messages
+        - That tool calls have distinct ids for a given assistant message
+        """
+        prev_role = None
+        expected_tool_ids: Set[str] = set()
+        observed_tool_ids: Set[str] = set()
+        for message in messages:
+            if prev_role is None:
+                prev_role = message.role
+                continue
+
+            if message.role == Roles.tool:
+                tool_call_id = message.tool_call_id
+                if tool_call_id in observed_tool_ids:
+                    raise InvalidMessageStructureException(f"Duplicate tool call id {tool_call_id} in tool results")
+                if tool_call_id not in expected_tool_ids:
+                    raise InvalidMessageStructureException(f"Unexpected tool call id {tool_call_id} in tool results")
+                observed_tool_ids.add(tool_call_id)
+
+            elif message.role == Roles.assistant:
+                # if we have an assistant message and we have not recieved all the function calls
+                # we need to raise an exception
+                if len(expected_tool_ids) != len(observed_tool_ids):
+                    raise InvalidMessageStructureException("Not the same number of function calls and responses")
+
+                expected_tool_ids.clear()
+                observed_tool_ids.clear()
+                if message.tool_calls is not None:
+                    # Validate that the number of function calls and ids are the same
+                    for tool_call in message.tool_calls:
+                        if tool_call.id in expected_tool_ids:
+                            raise InvalidMessageStructureException(
+                                f"Duplicate tool call id {tool_call.id} in assistant message"
+                            )
+                        expected_tool_ids.add(tool_call.id)
+
+            prev_role = message.role
+
+        if len(expected_tool_ids) != len(observed_tool_ids) and self._mode == ValidationMode.serving:
+            raise InvalidMessageStructureException("Not the same number of function calls and responses")
+        elif len(expected_tool_ids) < len(observed_tool_ids) and self._mode == ValidationMode.finetuning:
+            raise InvalidMessageStructureException("More tool responses than tool calls")
