@@ -4,11 +4,17 @@ from typing import Any, Dict, Generic, List, Optional, Tuple, Union
 
 import numpy as np
 
-from mistral_common.exceptions import TokenizerException
+from mistral_common.exceptions import (
+    InvalidAssistantMessageException,
+    InvalidMessageStructureException,
+    TokenizerException,
+)
 from mistral_common.protocol.instruct.messages import (
     AssistantMessage,
     AssistantMessageType,
     ContentChunk,
+    ImageChunk,
+    ImageURLChunk,
     SystemMessage,
     TextChunk,
     ToolMessage,
@@ -20,12 +26,14 @@ from mistral_common.tokens.tokenizers.base import (
     FIMRequestType,
     InstructRequestType,
     InstructTokenizer,
+    SpecialTokenPolicy,
     SpecialTokens,
     Tokenized,
     TokenizedType,
     Tokenizer,
+    UserMessagePosition,
 )
-from mistral_common.tokens.tokenizers.multimodal import MultiModalEncoder
+from mistral_common.tokens.tokenizers.image import ImageEncoder
 
 
 class InstructTokenizerBase(
@@ -33,16 +41,23 @@ class InstructTokenizerBase(
 ):
     r"""Base instruct tokenizer."""
 
-    def __init__(self, tokenizer: Tokenizer, mm_encoder: Optional[MultiModalEncoder] = None):
+    def __init__(self, tokenizer: Tokenizer, image_encoder: Optional[ImageEncoder] = None):
         r"""Initialize the instruct tokenizer.
 
         Args:
             tokenizer: The tokenizer to use.
-            mm_encoder: The multi-modal encoder to use if any.
+            image_encoder: The image encoder to use if any.
         """
         self.tokenizer = tokenizer
-        self.mm_encoder = mm_encoder
-        super().__init__(tokenizer, mm_encoder)
+        self.image_encoder = image_encoder
+        super().__init__(tokenizer, image_encoder)
+
+    @property
+    def mm_encoder(self) -> Optional[ImageEncoder]:
+        # this funtion is deprecated, use image_encoder instead
+        # TODO(Patrick) - throw a deprecation warning once
+        # changes applied to vllm and transformers
+        return self.image_encoder
 
     def start(self) -> List[int]:
         r"""Return the start tokens."""
@@ -77,7 +92,9 @@ class InstructTokenizerBase(
         raise NotImplementedError("Tool message not implemented")
 
     @abstractmethod
-    def encode_assistant_message(self, message: AssistantMessageType, is_before_last_user_message: bool) -> List[int]:
+    def encode_assistant_message(
+        self, message: AssistantMessageType, is_before_last_user_message: bool, continue_message: bool
+    ) -> List[int]:
         r"""Encode an assistant message.
 
         Raises:
@@ -115,6 +132,14 @@ class InstructTokenizerBase(
         # find last user message
         first_user_idx, last_user_idx = self.find_first_last_user(request)
         for msg_idx, msg in enumerate(request.messages):
+            if (
+                request.continue_final_message
+                and (msg_idx == len(request.messages) - 1)
+                and not isinstance(msg, AssistantMessage)
+            ):
+                raise InvalidMessageStructureException(
+                    "Cannot continue final message if it is not an assistant message"
+                )
             if isinstance(msg, UserMessage):
                 new_tokens, new_images = self.encode_user_message(
                     msg,
@@ -128,11 +153,17 @@ class InstructTokenizerBase(
             elif isinstance(msg, ToolMessage):
                 new_tokens = self.encode_tool_message(msg, msg_idx < last_user_idx)
             elif isinstance(msg, AssistantMessage):
-                new_tokens = self.encode_assistant_message(msg, msg_idx < last_user_idx)
+                continue_message = request.continue_final_message and (msg_idx == len(request.messages) - 1)
+
+                new_tokens = self.encode_assistant_message(
+                    msg, msg_idx < last_user_idx, continue_message=continue_message
+                )
                 if msg_idx == len(request.messages) - 1:
                     prefix_ids = new_tokens
             elif isinstance(msg, SystemMessage):
                 new_tokens = self.encode_system_message(msg)
+            else:
+                raise TokenizerException(f"Unknown message type {type(msg)}")
 
             tokens_list.append(new_tokens)
 
@@ -151,21 +182,30 @@ class InstructTokenizerBase(
 
         return Tokenized(
             tokens=tokens,
-            text=self.tokenizer.to_string(tokens),
+            text=self.decode(tokens, special_token_policy=SpecialTokenPolicy.KEEP),
             prefix_ids=prefix_ids,
             images=images,
         )
 
-    def decode(self, tokens: List[int]) -> str:
+    def decode(self, tokens: List[int], special_token_policy: Optional[SpecialTokenPolicy] = None) -> str:
         r"""Decode tokens to a string.
 
         Args:
             tokens: The tokens to decode.
+            special_token_policy: The policy to use for special tokens.
+                Passing `None` will default to `self._special_token_policy` for
+                [Tekkenizer][mistral_common.tokens.tokenizers.tekken.Tekkenizer] and `SpecialTokenPolicy.IGNORE`
+                for [SentencePieceTokenizer][mistral_common.tokens.tokenizers.sentencepiece.SentencePieceTokenizer].
+                Note that passing `None` will be deprecated and `special_token_policy` will default to
+                `SpecialTokenPolicy.IGNORE` in `mistral_common=1.7.0`.
 
         Returns:
             The decoded string.
         """
-        return self.tokenizer.decode(tokens)
+        return self.tokenizer.decode(tokens, special_token_policy=special_token_policy)
+
+    def _to_string(self, tokens: List[int]) -> str:
+        return self.tokenizer._to_string(tokens)
 
 
 class InstructTokenizerV1(
@@ -173,7 +213,7 @@ class InstructTokenizerV1(
 ):
     r"""Instruct tokenizer V1.
 
-    This tokenizer has basic for messages. It does not support tools or multi-modal inputs.
+    This tokenizer has basic for messages. It does not support tools or image inputs.
     """
 
     def encode_user_message(
@@ -198,9 +238,8 @@ class InstructTokenizerV1(
         Returns:
             The encoded tokens and empty list.
         """
-        assert message.content is not None
         assert isinstance(message.content, str), "Message content must be normalized"
-        assert self.mm_encoder is None, "InstructTokenizerV1 cannot encode images"
+        assert self.image_encoder is None, "InstructTokenizerV1 cannot encode images"
 
         content = ""
         if is_first and system_prompt:
@@ -246,12 +285,16 @@ class InstructTokenizerV1(
         """
         raise TokenizerException("Tools not implemented for tokenizer V1")
 
-    def encode_assistant_message(self, message: AssistantMessageType, is_before_last_user_message: bool) -> List[int]:
+    def encode_assistant_message(
+        self, message: AssistantMessageType, is_before_last_user_message: bool, continue_message: bool
+    ) -> List[int]:
         r"""Encode an assistant message.
 
         Args:
             message: The message to encode.
             is_before_last_user_message: Not used.
+            continue_message: Whether to continue the message generation.
+                Only use this if the assistant message is the last message.
 
         Returns:
             The encoded tokens.
@@ -259,11 +302,15 @@ class InstructTokenizerV1(
         assert isinstance(message, AssistantMessage), message
         if message.tool_calls is not None and len(message.tool_calls) > 0:
             raise TokenizerException("Tools not implemented for tokenizer V1")
+        if continue_message and message.prefix:
+            raise InvalidAssistantMessageException(
+                "`continue_message` is only supported for assistant messages that have `prefix=False`."
+            )
         elif message.content:
             curr_tokens = self.tokenizer.encode(message.content, bos=False, eos=False)
         else:
             raise TokenizerException(f"{message.content} // {message.tool_calls}")
-        if not message.prefix:
+        if not message.prefix and not continue_message:
             curr_tokens.append(self.tokenizer.eos_id)
         return curr_tokens
 
@@ -284,14 +331,16 @@ class InstructTokenizerV2(
     This tokenizer adds supports to images, tools and FIM requests.
     """
 
-    def __init__(self, tokenizer: Tokenizer, mm_encoder: Optional[MultiModalEncoder] = None):
+    _user_message_position_to_encode_tools = UserMessagePosition.last
+
+    def __init__(self, tokenizer: Tokenizer, image_encoder: Optional[ImageEncoder] = None):
         r"""Initialize the tokenizer.
 
         Args:
             tokenizer: The tokenizer to use.
-            mm_encoder: The multi-modal encoder to use.
+            image_encoder: The image encoder to use.
         """
-        super().__init__(tokenizer, mm_encoder)
+        super().__init__(tokenizer, image_encoder)
         self.BEGIN_INST = self.tokenizer.get_control_token(SpecialTokens.begin_inst.value)
         self.END_INST = self.tokenizer.get_control_token(SpecialTokens.end_inst.value)
         self.BEGIN_AVAILABLE_TOOLS = self.tokenizer.get_control_token(SpecialTokens.begin_tools.value)
@@ -325,9 +374,12 @@ class InstructTokenizerV2(
         Returns:
             The encoded tokens and the list of images.
         """
-        assert message.content is not None
+        do_encode_tools = False
+        do_encode_tools |= is_first and (self._user_message_position_to_encode_tools == UserMessagePosition.first)
+        do_encode_tools |= is_last and (self._user_message_position_to_encode_tools == UserMessagePosition.last)
         tools_tokens: List[int] = []
-        if is_last and available_tools:
+
+        if do_encode_tools and available_tools:
             tools = [tool.model_dump() for tool in available_tools]
             tools_json_tokens = self.tokenizer.encode(json.dumps(tools, ensure_ascii=False), bos=False, eos=False)
             tools_tokens = [
@@ -358,7 +410,6 @@ class InstructTokenizerV2(
 
     def _prepare_tool_result(self, tool_message: ToolMessage) -> Dict[str, Any]:
         r"""Bit of a hack due to the way tool results are tokenized."""
-        assert tool_message.content is not None, "Tool message content cannot be None"
         return {
             "name": tool_message.name,
             "content": self._parse_json_content(tool_message.content),
@@ -397,6 +448,7 @@ class InstructTokenizerV2(
 
     def _encode_normal_content_assistant_message(self, message: AssistantMessageType) -> List[int]:
         assert message.content, f"Assistant message must have content. Got {message}"
+        assert isinstance(message.content, str), "Message content must be a string for tokenizer < V7"
         return self.tokenizer.encode(message.content.rstrip(" "), bos=False, eos=False)
 
     def _encode_tool_calls_in_assistant_message(self, message: AssistantMessageType) -> List[int]:
@@ -411,19 +463,27 @@ class InstructTokenizerV2(
         ]
         return curr_tokens
 
-    def encode_assistant_message(self, message: AssistantMessageType, is_before_last_user_message: bool) -> List[int]:
+    def encode_assistant_message(
+        self, message: AssistantMessageType, is_before_last_user_message: bool, continue_message: bool
+    ) -> List[int]:
         r"""Encode an assistant message.
 
         Args:
             message: The message to encode.
             is_before_last_user_message: Whether the message is before the last user message. If has tools and true, the
                 message is not encoded.
+            continue_message: Whether to continue the message generation.
+                Only use this if the assistant message is the last message.
 
         Returns:
             The encoded tokens.
         """
         if message.tool_calls and message.content:
             raise ValueError(f"Cannot have tool calls and content defined in the same assistant message {message}")
+        if continue_message and message.prefix:
+            raise InvalidAssistantMessageException(
+                "`continue_message` is only supported for assistant messages that have `prefix=False`."
+            )
 
         if message.tool_calls:
             if is_before_last_user_message:
@@ -434,7 +494,7 @@ class InstructTokenizerV2(
             curr_tokens = self._encode_normal_content_assistant_message(message)
         else:
             raise TokenizerException(f"Invalid assistant message: {message.content}")
-        if not message.prefix:
+        if not message.prefix and not continue_message:
             curr_tokens.append(self.tokenizer.eos_id)
         return curr_tokens
 
@@ -461,7 +521,7 @@ class InstructTokenizerV2(
             self.PREFIX,
             *prefix_tokens,
         ]
-        return Tokenized(tokens=tokens, text=self.tokenizer.to_string(tokens))
+        return Tokenized(tokens=tokens, text=self.decode(tokens, special_token_policy=SpecialTokenPolicy.KEEP))
 
 
 class InstructTokenizerV3(
@@ -472,14 +532,14 @@ class InstructTokenizerV3(
     The only difference with V2 tokenizer is that it encodes the tool messages differently.
     """
 
-    def __init__(self, tokenizer: Tokenizer, mm_encoder: Optional[MultiModalEncoder] = None) -> None:
+    def __init__(self, tokenizer: Tokenizer, image_encoder: Optional[ImageEncoder] = None) -> None:
         r"""Initialize the tokenizer.
 
         Args:
             tokenizer: The tokenizer to use.
-            mm_encoder: The multi-modal encoder to use.
+            image_encoder: The image encoder to use.
         """
-        super().__init__(tokenizer, mm_encoder=mm_encoder)
+        super().__init__(tokenizer, image_encoder=image_encoder)
 
     def _prepare_function_call(self, tool_call: ToolCall) -> Dict[str, Any]:
         function_call = {
@@ -493,7 +553,6 @@ class InstructTokenizerV3(
         return function_call
 
     def _prepare_tool_result(self, tool_message: ToolMessage) -> Dict[str, Any]:
-        assert tool_message.content is not None, "Tool message content cannot be None"
         assert tool_message.tool_call_id is not None, "Tool message has to have the tool call id defined in v3"
 
         return {
@@ -524,12 +583,16 @@ class InstructTokenizerV3(
         ]
         return curr_tokens
 
-    def encode_assistant_message(self, message: AssistantMessageType, is_before_last_user_message: bool) -> List[int]:
+    def encode_assistant_message(
+        self, message: AssistantMessageType, is_before_last_user_message: bool, continue_message: bool
+    ) -> List[int]:
         r"""Encode an assistant message.
 
         Note:
             Same as [V2][mistral_common.tokens.tokenizers.instruct.InstructTokenizerV2.encode_assistant_message] but
             always encode the tool history.
+            continue_message: Whether to continue the message generation.
+                Only use this if the assistant message is the last message.
 
         Args:
             message: The message to encode.
@@ -538,7 +601,7 @@ class InstructTokenizerV3(
         Returns:
             The encoded tokens.
         """
-        return super().encode_assistant_message(message, False)
+        return super().encode_assistant_message(message, False, continue_message)
 
     def encode_user_content(
         self,
@@ -565,7 +628,9 @@ class InstructTokenizerV3(
         images: List[np.ndarray] = []
 
         has_one_img_one_text_first = (
-            len(content) == 2 and isinstance(content[0], TextChunk) and not isinstance(content[1], TextChunk)
+            len(content) == 2
+            and isinstance(content[0], TextChunk)
+            and isinstance(content[1], (ImageURLChunk, ImageChunk))
         )
         if force_img_first and has_one_img_one_text_first:
             # make sure that if exactly one image and text chunk are passed we force the image chunk to be first
@@ -581,11 +646,11 @@ class InstructTokenizerV3(
                 content += chunk.text
                 tokens.extend(self.tokenizer.encode(content, bos=False, eos=False))
             else:
-                assert self.mm_encoder is not None, "Make sure to define a multi-modal encoder at init"
+                assert self.image_encoder is not None, "Make sure to define a image encoder at init"
                 if content:
                     tokens.extend(self.tokenizer.encode(content, bos=False, eos=False))
 
-                img_encoding = self.mm_encoder(chunk)
+                img_encoding = self.image_encoder(chunk)
 
                 tokens.extend(img_encoding.tokens)
                 images.append(img_encoding.image)
@@ -603,15 +668,15 @@ class InstructTokenizerV7(InstructTokenizerV3):
 
     """
 
-    def __init__(self, tokenizer: Tokenizer, mm_encoder: Optional[MultiModalEncoder] = None) -> None:
+    def __init__(self, tokenizer: Tokenizer, image_encoder: Optional[ImageEncoder] = None) -> None:
         r"""Initialize the tokenizer.
 
         Args:
             tokenizer: The tokenizer to use.
-            mm_encoder: The multi-modal encoder to use.
+            image_encoder: The image encoder to use.
         """
 
-        super().__init__(tokenizer, mm_encoder)
+        super().__init__(tokenizer, image_encoder)
         self.BEGIN_SYSTEM = self.tokenizer.get_control_token(SpecialTokens.begin_system.value)
         self.END_SYSTEM = self.tokenizer.get_control_token(SpecialTokens.end_system.value)
         self.BEGIN_TOOL_CONTENT = self.tokenizer.get_control_token(SpecialTokens.begin_tool_content.value)
@@ -666,8 +731,6 @@ class InstructTokenizerV7(InstructTokenizerV3):
         Returns:
             The encoded tokens.
         """
-        assert message.content is not None
-        assert isinstance(message.content, str), "Message content must be normalized"
         tokens = [
             self.BEGIN_SYSTEM,
             *self.tokenizer.encode(message.content, bos=False, eos=False),
@@ -722,6 +785,7 @@ class InstructTokenizerV7(InstructTokenizerV3):
             The encoded tokens.
         """
         assert message.tool_call_id is not None
+        assert isinstance(message.content, str), "Message content must be normalized"
         tool_call_id_tokens = self.tokenizer.encode(message.tool_call_id, bos=False, eos=False)
         tokens = self.tokenizer.encode(message.content, bos=False, eos=False)
 
@@ -737,29 +801,33 @@ class InstructTokenizerV7(InstructTokenizerV3):
         ]
         return curr_tokens
 
-    def encode_assistant_message(self, message: AssistantMessageType, is_before_last_user_message: bool) -> List[int]:
+    def encode_assistant_message(
+        self, message: AssistantMessageType, is_before_last_user_message: bool, continue_message: bool
+    ) -> List[int]:
         r"""Encode an assistant message.
 
         Args:
             message: The message to encode.
             is_before_last_user_message: Not used.
+            continue_message: Whether to continue the message generation.
+                Only use this if the assistant message is the last message.
 
         Returns:
             The encoded tokens.
         """
         if not message.content and not message.tool_calls:
             raise TokenizerException(f"Invalid assistant message: {message}")
+        if continue_message and message.prefix:
+            raise InvalidAssistantMessageException(
+                "`continue_message` is only supported for assistant messages that have `prefix=False`."
+            )
+
         curr_tokens: list = []
         if message.content:
-            if isinstance(message.content, str):
-                curr_tokens += self._encode_normal_content_assistant_message(message)
-            elif isinstance(message.content, list):
-                curr_tokens += self.encode_content_chunks(
-                    message.content, is_last=False, system_prompt=None, force_img_first=True
-                ).tokens
+            curr_tokens += self._encode_normal_content_assistant_message(message)
         if message.tool_calls:
             curr_tokens += self._encode_tool_calls_in_assistant_message(message)
-        if not message.prefix:
+        if not message.prefix and not continue_message:
             curr_tokens.append(self.tokenizer.eos_id)
 
         return curr_tokens
@@ -776,9 +844,9 @@ class InstructTokenizerV11(InstructTokenizerV7):
     def __init__(
         self,
         tokenizer: Tokenizer,
-        mm_encoder: Optional[MultiModalEncoder] = None,
+        image_encoder: Optional[ImageEncoder] = None,
     ) -> None:
-        super().__init__(tokenizer, mm_encoder)
+        super().__init__(tokenizer, image_encoder)
         self.ARGS = self.tokenizer.get_control_token(SpecialTokens.args.value)
         self.CALL_ID = self.tokenizer.get_control_token(SpecialTokens.call_id.value)
 
@@ -800,3 +868,48 @@ class InstructTokenizerV11(InstructTokenizerV7):
                 *self.tokenizer.encode(json.dumps(prepared["arguments"], ensure_ascii=False), bos=False, eos=False),
             ]
         return curr_tokens
+
+
+class InstructTokenizerV13(InstructTokenizerV11):
+    r"""Instruct tokenizer V13.
+
+    The difference with V11 tokenizer is that it encodes tool calls differently:
+        - available tools are tokenized at the first user message.
+        - call id is no longer tokenized for tool calls or results.
+    """
+
+    _user_message_position_to_encode_tools = UserMessagePosition.first
+
+    def _encode_tool_calls_in_assistant_message(self, message: AssistantMessageType) -> List[int]:
+        assert message.tool_calls, f"Assistant message must have tool calls. Got {message}"
+        curr_tokens = []
+        for tool_call in message.tool_calls:
+            assert tool_call.id and tool_call.id != "null"
+            prepared = self._prepare_function_call(tool_call)
+
+            curr_tokens += [
+                self.TOOL_CALLS,
+                *self.tokenizer.encode(prepared["name"], bos=False, eos=False),
+                self.ARGS,
+                *self.tokenizer.encode(json.dumps(prepared["arguments"], ensure_ascii=False), bos=False, eos=False),
+            ]
+        return curr_tokens
+
+    def encode_tool_message(self, message: ToolMessage, is_before_last_user_message: bool) -> List[int]:
+        r"""Encode a tool message.
+
+        Args:
+            message: The message to encode.
+            is_before_last_user_message: Not used.
+        Returns:
+            The encoded tokens.
+        """
+        assert message.tool_call_id is not None, "Tool call id must be provided for tokenizer >= v13"
+
+        tokens = self.tokenizer.encode(message.content, bos=False, eos=False)
+        curr_tokens = [
+            self.BEGIN_TOOL_RESULTS,
+            *tokens,
+            self.END_TOOL_RESULTS,
+        ]
+        return curr_tokens
