@@ -2,7 +2,6 @@ import base64
 import json
 import logging
 import warnings
-from enum import Enum
 from functools import cached_property
 from itertools import groupby
 from pathlib import Path
@@ -11,11 +10,12 @@ from typing import Dict, List, Optional, Type, TypedDict, Union
 import tiktoken
 
 from mistral_common.tokens.tokenizers.base import (
+    SpecialTokenPolicy,
     SpecialTokens,
     Tokenizer,
     TokenizerVersion,
 )
-from mistral_common.tokens.tokenizers.multimodal import MultimodalConfig
+from mistral_common.tokens.tokenizers.image import ImageConfig
 
 logger = logging.getLogger(__name__)
 
@@ -82,7 +82,7 @@ class ModelData(TypedDict):
         config: The configuration of the tokenizer.
         version: The version of the tokenizer.
         type: The type of the tokenizer.
-        multimodal: The multimodal configuration of the tokenizer.
+        image: The image configuration of the tokenizer.
     """
 
     vocab: List[TokenInfo]
@@ -90,21 +90,7 @@ class ModelData(TypedDict):
     config: TekkenConfig
     version: int
     type: str
-    multimodal: MultimodalConfig
-
-
-class SpecialTokenPolicy(Enum):
-    r"""What to do with special tokens when encoding/decoding.
-
-    Attributes:
-        IGNORE: Ignore special tokens.
-        KEEP: Keep special tokens.
-        RAISE: Raise an error if special tokens are found.
-    """
-
-    IGNORE = 0
-    KEEP = 1
-    RAISE = 2
+    image: ImageConfig
 
 
 class Tekkenizer(Tokenizer):
@@ -151,8 +137,8 @@ class Tekkenizer(Tokenizer):
         version: TokenizerVersion,
         *,
         name: str = "tekkenizer",
-        _path: Optional[str] = None,
-        mm_config: Optional[MultimodalConfig] = None,
+        _path: Optional[Union[str, Path]] = None,
+        image_config: Optional[ImageConfig] = None,
     ):
         r"""Initialize the tekken tokenizer.
 
@@ -164,7 +150,7 @@ class Tekkenizer(Tokenizer):
             num_special_tokens: The number of special tokens of the tokenizer.
             version: The version of the tokenizer.
             name: The name of the tokenizer.
-            mm_config: The multimodal configuration of the tokenizer.
+            image_config: The image configuration of the tokenizer.
         """
         assert vocab_size <= len(vocab) + num_special_tokens, (
             vocab_size,
@@ -208,11 +194,19 @@ class Tekkenizer(Tokenizer):
         )
 
         self._version = version
-        self._mm_config = mm_config
+        self._image_config = image_config
         self._all_special_tokens = special_tokens
         self._special_tokens_reverse_vocab = {t["token_str"]: t["rank"] for t in special_tokens}
         self._vocab = [self.id_to_piece(i) for i in range(vocab_size)]
         self._special_token_policy = SpecialTokenPolicy.IGNORE
+        self._file_path = Path(_path) if _path is not None else None
+
+    @property
+    def file_path(self) -> Path:
+        r"""The path to the tokenizer file."""
+        if self._file_path is None:
+            raise ValueError("The tokenizer was not loaded from a file.")
+        return self._file_path
 
     @classmethod
     def from_file(cls: Type["Tekkenizer"], path: Union[str, Path]) -> "Tekkenizer":
@@ -227,7 +221,7 @@ class Tekkenizer(Tokenizer):
         if isinstance(path, str):
             path = Path(path)
         assert path.exists(), path
-        with open(path, "r") as f:
+        with open(path, "r", encoding="utf-8") as f:
             untyped = json.load(f)
 
         _version_str = untyped["config"].get("version")
@@ -242,19 +236,13 @@ class Tekkenizer(Tokenizer):
 
         special_tokens_dicts: Optional[List[SpecialTokenInfo]] = untyped.get("special_tokens", None)
         if special_tokens_dicts is None:
-            err_msg = (
-                f"Special tokens not found in {path} and default to {Tekkenizer.DEPRECATED_SPECIAL_TOKENS}. "
-                "This behavior will be deprecated going forward. "
-                "Please update your tokenizer file and include all special tokens you need."
-            )
             # Tokenizer > v7 should find special tokens in the tokenizer file
             if version > TokenizerVersion("v7"):
-                raise ValueError(err_msg)
-            else:
-                warnings.warn(
-                    err_msg,
-                    FutureWarning,
+                raise ValueError(
+                    f"Special tokens not found in {path}. "
+                    "Please update your tokenizer file and include all special tokens you need."
                 )
+            else:
                 special_tokens = list(Tekkenizer.DEPRECATED_SPECIAL_TOKENS)
         else:
             special_tokens = [token for token in special_tokens_dicts]
@@ -262,7 +250,15 @@ class Tekkenizer(Tokenizer):
         untyped["special_tokens"] = special_tokens
 
         if mm := untyped.get("multimodal", None):
-            untyped["multimodal"] = MultimodalConfig(**mm)
+            # deprecated - only allowed for tokenizers <= v11
+            if version > TokenizerVersion("v11"):
+                raise ValueError(
+                    f"The image config has to be called 'image' in {path} for tokenizers of version {version.value}."
+                )
+
+            untyped["image"] = ImageConfig(**mm)
+        elif image := untyped.get("image", None):
+            untyped["image"] = ImageConfig(**image)
 
         model_data: ModelData = untyped
 
@@ -274,17 +270,18 @@ class Tekkenizer(Tokenizer):
             num_special_tokens=model_data["config"]["default_num_special_tokens"],
             version=version,
             name=path.name.replace(".json", ""),
-            mm_config=model_data.get("multimodal"),
+            image_config=model_data.get("image"),
+            _path=path,
         )
 
     @property
-    def multimodal(self) -> Optional[MultimodalConfig]:
-        r"""The multimodal configuration of the tokenizer."""
-        return self._mm_config
+    def image(self) -> Optional[ImageConfig]:
+        r"""The image configuration of the tokenizer."""
+        return self._image_config
 
-    @multimodal.setter
-    def multimodal(self, value: MultimodalConfig) -> None:
-        raise ValueError("Can only set Multimodal config at init")
+    @image.setter
+    def image(self, value: ImageConfig) -> None:
+        raise ValueError("Can only set Image config at init")
 
     @property
     def num_special_tokens(self) -> int:
@@ -308,6 +305,18 @@ class Tekkenizer(Tokenizer):
 
     @special_token_policy.setter
     def special_token_policy(self, policy: SpecialTokenPolicy) -> None:
+        r"""Set the policy for handling special tokens."""
+        if not isinstance(policy, SpecialTokenPolicy):
+            raise ValueError(f"Expected SpecialTokenPolicy, got {type(policy)}.")
+
+        warnings.warn(
+            (
+                "The attributed `special_token_policy` is deprecated and will be removed in 1.7.0. "
+                "Please pass a special token policy explicitly to the relevant methods."
+            ),
+            FutureWarning,
+        )
+
         self._special_token_policy = policy
 
     @cached_property
@@ -405,25 +414,95 @@ class Tekkenizer(Tokenizer):
         else:
             raise ValueError(f"Unknown control token {s}")
 
-    def decode(self, tokens: List[int]) -> str:
-        r"""Decode a list of token ids into a string."""
-        return "".join(self._decode_all(tokens, special_token_policy=self._special_token_policy))
+    def decode(self, tokens: List[int], special_token_policy: Optional[SpecialTokenPolicy] = None) -> str:
+        r"""Decode a list of token ids into a string.
+
+        Args:
+            tokens: The list of token ids to decode.
+            special_token_policy: The policy for handling special tokens.
+                Use the tokenizer's [attribute][mistral_common.tokens.tokenizers.tekken.Tekkenizer.special_token_policy]
+                if `None`. Passing `None` is deprecated and will be changed
+                to `SpecialTokenPolicy.IGNORE` in `mistral_common=1.7.0`.
+
+        Returns:
+            The decoded string.
+        """
+        if special_token_policy is not None and not isinstance(special_token_policy, SpecialTokenPolicy):
+            raise ValueError(
+                f"Expected `special_token_policy` to be None or SpecialTokenPolicy, got {type(special_token_policy)}."
+            )
+
+        if special_token_policy is None:
+            warnings.warn(
+                (
+                    f"Using the tokenizer's special token policy ({self._special_token_policy}) is deprecated. "
+                    "It will be removed in 1.7.0. "
+                    "Please pass a special token policy explicitly. "
+                    "Future default will be SpecialTokenPolicy.IGNORE."
+                ),
+                FutureWarning,
+            )
+            special_token_policy = self._special_token_policy
+
+        return "".join(self._decode_all(tokens, special_token_policy=special_token_policy))
 
     def to_string(self, tokens: List[int]) -> str:
-        r"""Decode a list of token ids into a string keeping special tokens for debugging purposes."""
-        return "".join(self._decode_all(tokens, special_token_policy=SpecialTokenPolicy.KEEP))
+        r"""[DEPRECATED] Converts a list of token ids into a string, keeping special tokens.
+
+        Use `decode` with `special_token_policy=SpecialTokenPolicy.KEEP` instead.
+
+        This is a convenient method for debugging.
+        """
+        warnings.warn(
+            (
+                "`to_string` is deprecated and will be removed in 1.7.0. "
+                "Use `decode` with `special_token_policy=SpecialTokenPolicy.KEEP` instead."
+            ),
+            FutureWarning,
+        )
+        return self._to_string(tokens)
+
+    def _to_string(self, tokens: List[int]) -> str:
+        return self.decode(tokens, special_token_policy=SpecialTokenPolicy.KEEP)
 
     def id_to_piece(self, token_id: int) -> str:
         r"""Convert a token id to its string representation."""
-        return self._decode_all([token_id], special_token_policy=SpecialTokenPolicy.KEEP)[0]
+        return self.decode([token_id], special_token_policy=SpecialTokenPolicy.KEEP)
+
+    def id_to_byte_piece(self, token_id: int, special_token_policy: Optional[SpecialTokenPolicy] = None) -> bytes:
+        r"""Convert a token id to its byte representation.
+
+        Args:
+            token_id: The token id to convert.
+            special_token_policy: The policy for handling special tokens.
+                Use the tokenizer's [attribute][mistral_common.tokens.tokenizers.tekken.Tekkenizer.special_token_policy]
+                if `None`. Passing `None` is deprecated and will be changed
+                to `SpecialTokenPolicy.IGNORE` in `mistral_common=1.7.0`.
+
+        Returns:
+            The byte representation of the token.
+        """
+        if special_token_policy is None:
+            warnings.warn(
+                (
+                    f"Using the tokenizer's special token policy ({self._special_token_policy}) is deprecated. "
+                    "It will be removed in 1.7.0. "
+                    "Please pass a special token policy explicitly. "
+                    "Future default will be SpecialTokenPolicy.IGNORE."
+                ),
+                FutureWarning,
+            )
+            special_token_policy = self._special_token_policy
 
-    def id_to_byte_piece(self, token_id: int) -> bytes:
-        r"""Convert a token id to its byte representation."""
         if token_id < self.num_special_tokens:
-            if self._special_token_policy == SpecialTokenPolicy.KEEP:
+            if special_token_policy == SpecialTokenPolicy.KEEP:
                 return self._all_special_tokens[token_id]["token_str"].encode("utf-8")
-            elif self._special_token_policy == SpecialTokenPolicy.RAISE:
+            elif special_token_policy == SpecialTokenPolicy.RAISE:
                 raise ValueError(f"{token_id} is a special token")
+            elif special_token_policy == SpecialTokenPolicy.IGNORE:
+                return b""
+            else:
+                raise ValueError(f"Unknown special token policy {special_token_policy}")
 
         return self._model.decode_single_token_bytes(token_id - self.num_special_tokens)
 
