@@ -1,5 +1,5 @@
 import json
-from typing import Dict, List
+from typing import Dict, List, Union
 
 import pytest
 
@@ -7,6 +7,7 @@ from mistral_common.protocol.instruct.messages import (
     AssistantMessage,
     ChatMessage,
     ChunkTypes,
+    ContentChunk,
     FinetuningAssistantMessage,
     FinetuningMessage,
     SystemMessage,
@@ -14,7 +15,11 @@ from mistral_common.protocol.instruct.messages import (
     ToolMessage,
     UserMessage,
 )
-from mistral_common.protocol.instruct.normalize import InstructRequestNormalizer, InstructRequestNormalizerV7
+from mistral_common.protocol.instruct.normalize import (
+    InstructRequestNormalizer,
+    InstructRequestNormalizerV7,
+    InstructRequestNormalizerV13,
+)
 from mistral_common.protocol.instruct.request import ChatCompletionRequest
 from mistral_common.protocol.instruct.tool_calls import Function, FunctionCall, Tool, ToolCall
 from mistral_common.tokens.instruct.request import InstructRequest
@@ -121,7 +126,7 @@ class TestChatCompletionRequestNormalization:
         self,
         roles: List[str],
         expected_roles: List[str],
-        expected_content: List[str],
+        expected_content: List[Union[List[ContentChunk], str]],
         normalizer: InstructRequestNormalizer,
     ) -> None:
         letter_to_cls: Dict[str, ChatMessage] = {
@@ -218,26 +223,9 @@ class TestChatCompletionRequestNormalization:
         assert first_message.content == "user"
         assert parsed_request.system_prompt == "system"
 
-    def test_system_prompt_chunks_aggregated(self, normalizer: InstructRequestNormalizer) -> None:
-        chat_completion_request = mock_chat_completion(
-            messages=[
-                UserMessage(content="foo"),
-                SystemMessage(
-                    content=[
-                        TextChunk(type=ChunkTypes.text, text="chunk1"),
-                        TextChunk(type=ChunkTypes.text, text="chunk2"),
-                        TextChunk(type=ChunkTypes.text, text="chunk3"),
-                    ],
-                ),
-            ],
-        )
-
-        parsed_request = normalizer.from_chat_completion_request(chat_completion_request)
-        assert parsed_request.system_prompt == "chunk1\n\nchunk2\n\nchunk3"
-
     def test_normalize_tools(self, normalizer: InstructRequestNormalizer) -> None:
         """
-        Test doesnt really "normalize" anything but it checks that the tools are added to the
+        Test doesn't really "normalize" anything but it checks that the tools are added to the
         InstructRequest during from_chat_completion_request
         """
         tools = [
@@ -466,3 +454,190 @@ class TestFineTuningNormalizer:
         )
         normalized = normalizer.from_chat_completion_request(request)
         assert normalized == expected
+
+
+class TestChatCompletionRequestNormalizationV13:
+    @pytest.fixture(autouse=True)
+    def normalizer_v13(self) -> InstructRequestNormalizerV13:
+        return InstructRequestNormalizerV13(UserMessage, AssistantMessage, ToolMessage, SystemMessage, InstructRequest)
+
+    def _mock_chat_completion(self, messages: List[ChatMessage]) -> ChatCompletionRequest:
+        return ChatCompletionRequest(
+            model="test",
+            messages=messages,
+            top_p=1.0,
+            temperature=0.7,
+        )
+
+    def test_no_reorder_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
+        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
+            messages=[
+                UserMessage(content="A"),
+                AssistantMessage(
+                    content="B",
+                    tool_calls=[
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="C", tool_call_id="1"),
+                ToolMessage(content="D", tool_call_id="2"),
+            ]
+        )
+        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
+            chat_completion_request
+        )
+        assert parsed_request.messages == [
+            UserMessage(content="A"),
+            AssistantMessage(
+                content="B",
+                tool_calls=[
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="C", tool_call_id="1"),
+            ToolMessage(content="D", tool_call_id="2"),
+        ]
+
+    def test_reorder_last_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
+        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
+            messages=[
+                UserMessage(content="A"),
+                AssistantMessage(
+                    content="B",
+                    tool_calls=[
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="D", tool_call_id="2"),
+                ToolMessage(content="C", tool_call_id="1"),
+            ]
+        )
+        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
+            chat_completion_request
+        )
+        assert parsed_request.messages == [
+            UserMessage(content="A"),
+            AssistantMessage(
+                content="B",
+                tool_calls=[
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="C", tool_call_id="1"),
+            ToolMessage(content="D", tool_call_id="2"),
+        ]
+
+    def test_reorder_internal_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
+        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
+            messages=[
+                UserMessage(content="A"),
+                AssistantMessage(
+                    content="B",
+                    tool_calls=[
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="D", tool_call_id="2"),
+                ToolMessage(content="C", tool_call_id="1"),
+                AssistantMessage(content="E"),
+            ]
+        )
+        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
+            chat_completion_request
+        )
+        assert parsed_request.messages == [
+            UserMessage(content="A"),
+            AssistantMessage(
+                content="B",
+                tool_calls=[
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="C", tool_call_id="1"),
+            ToolMessage(content="D", tool_call_id="2"),
+            AssistantMessage(content="E"),
+        ]
+
+    def test_reorder_extra_tool_messages(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
+        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
+            messages=[
+                UserMessage(content="A"),
+                AssistantMessage(
+                    content="B",
+                    tool_calls=[
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="D", tool_call_id="2"),
+                ToolMessage(content="C", tool_call_id="1"),
+            ]
+        )
+        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
+            chat_completion_request
+        )
+        assert parsed_request.messages == [
+            UserMessage(content="A"),
+            AssistantMessage(
+                content="B",
+                tool_calls=[
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="C", tool_call_id="1"),
+            ToolMessage(content="D", tool_call_id="2"),
+        ]
+
+    def test_reorder_only_from_latest_assistant_message(self, normalizer_v13: InstructRequestNormalizerV13) -> None:
+        chat_completion_request: ChatCompletionRequest = self._mock_chat_completion(
+            messages=[
+                UserMessage(content="A"),
+                AssistantMessage(
+                    content="B",
+                    tool_calls=[
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="C", tool_call_id="1"),
+                ToolMessage(content="D", tool_call_id="2"),
+                AssistantMessage(
+                    content="E",
+                    tool_calls=[
+                        ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                        ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ],
+                ),
+                ToolMessage(content="C", tool_call_id="1"),
+                ToolMessage(content="D", tool_call_id="2"),
+            ]
+        )
+        parsed_request: InstructRequest[ChatMessage, Tool] = normalizer_v13.from_chat_completion_request(
+            chat_completion_request
+        )
+        assert parsed_request.messages == [
+            UserMessage(content="A"),
+            AssistantMessage(
+                content="B",
+                tool_calls=[
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                    ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="C", tool_call_id="1"),
+            ToolMessage(content="D", tool_call_id="2"),
+            AssistantMessage(
+                content="E",
+                tool_calls=[
+                    ToolCall(id="2", function=FunctionCall(name="foo", arguments="{}")),
+                    ToolCall(id="1", function=FunctionCall(name="foo", arguments="{}")),
+                ],
+            ),
+            ToolMessage(content="D", tool_call_id="2"),
+            ToolMessage(content="C", tool_call_id="1"),
+        ]
